package org.helioviewer.jhv.internal_plugins.filter.SOHOLUTFilterPlugin;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.Map;
import java.util.TreeMap;

import org.helioviewer.base.FileUtils;
import org.helioviewer.base.logging.Log;
import org.helioviewer.jhv.JHVDirectory;

/**
 * Representing a single color lookup. Each lookup has a name assigned and gives
 * access to lookup the values
 * 
 * @author Helge Dietert
 */
public class LUT {

    public final static char angstrom = '\u212B';

    /**
     * Giving a name of the lut
     */
    private String lutName;

    /**
     * Returns the name of the lut
     * 
     * @return Name
     */
    public String getName() {
        return lutName;
    }

    /**
     * LUT for 8 bit
     */
    private int[] lut8;
    /**
     * LUT for 16 bit
     */
    private int[] lut16;

    /**
     * Creates a lookup table with the given table, use e.g. for int[]
     * org.helioviewer.viewmodel.view.jp2view.JHVJP2View.getBuiltInLUT()
     * 
     * @param name
     *            Name of the lookup
     * @param lookup8
     *            Array of lookup values for 8 bit lookup
     * @param lookup16
     *            Array of lookup values for 16 bit lookup
     */
    public LUT(String name, int[] lookup8, int[] lookup16) {
        lutName = name;
        this.lut8 = lookup8;
        this.lut16 = lookup16;
    }

    /**
     * Fills the output array with values based on the given input data using
     * the given lookup table.
     * 
     * The output data must have the same size as the input data to call this
     * function.
     * 
     * @param byteBuf
     *            Input data
     * @param intBuf
     *            Output data
     * @param invert
     *            If true, the color table is inverted, false otherwise
     */
    public void lookup8(byte[] byteBuf, int[] intBuf, boolean invert) {
        if (invert) {
            int offset = lut8.length - 1;
            for (int i = 0; i < byteBuf.length; ++i) {
                intBuf[i] = lut8[offset - (byteBuf[i] & 0xff)];
            }
        } else {
            for (int i = 0; i < byteBuf.length; ++i) {
                intBuf[i] = lut8[byteBuf[i] & 0xff];
            }
        }
    }

    /**
     * Fills the output array with values based on the given input data using
     * the given lookup table.
     * 
     * The output data must have the same size as the input data to call this
     * function.
     * 
     * @param shortBuf
     *            Input data
     * @param intBuf
     *            Output data
     * @param invert
     *            If true, the color table is inverted, false otherwise
     */
    public void lookup16(short[] shortBuf, int[] intBuf, boolean invert) {
        if (invert) {
            int offset = lut16.length - 1;
            for (int i = 0; i < shortBuf.length; ++i) {
                intBuf[i] = lut16[offset - (shortBuf[i] & 0xFFFF)];
            }
        } else {
            for (int i = 0; i < shortBuf.length; ++i) {
                intBuf[i] = lut16[(shortBuf[i] & 0xFFFF)];
            }
        }
    }

    /**
     * Gives back LUT for 8bit
     * 
     * @return table
     */
    public int[] getLut8() {
        return lut8;
    }

    /**
     * Gives back LUT for 16bit
     * 
     * @return table
     */
    public int[] getLut16() {
        return lut16;
    }

    /**
     * Reads in a gimp gradient file as a gradient filter
     * 
     * @param file
     * @return Created LUT
     * @throws GradientError
     *             Error parsing/evaluating the file
     * @throws IOException
     *             From opening the file
     * @throws FileNotFoundException
     *             From opening the file
     */
    public static LUT readGimpGradientFile(File file) throws GradientError, IOException, FileNotFoundException {
        FileInputStream fr = null;
        LUT l;
        try {
            fr = new FileInputStream(file);
            l = readGimpGradientStream(fr);
        } finally {
            fr.close();
        }
        return l;
    }

    /**
     * Reads in a gimp gradient file from a input stream as a gradient filter
     * 
     * @param is
     *            Input stream to read from
     * @return Created LUT
     * @throws GradientError
     *             Error parsing/evaluating the file
     * @throws IOException
     *             From opening the file
     */
    public static LUT readGimpGradientStream(InputStream is) throws GradientError, IOException {
        BufferedReader in = new BufferedReader(new InputStreamReader(is, "UTF-8"));
        GimpGradient gg = new GimpGradient(in);
        int[] lut8 = new int[256];
        int[] lut16 = new int[65536];
        int i;
        for (i = 0; i < 256; i++) {
            lut8[i] = gg.getGradientColor(i / 255.0);
        }
        for (i = 0; i < 65536; i++) {
            lut16[i] = gg.getGradientColor(i / 65535.0);
        }
        return new LUT(gg.getName(), lut8, lut16);
    }

    /**
     * List of shown color table maps, will be initialized with the first start
     * up
     */
    private static Map<String, LUT> standardList = null;

    /**
     * Creates a standard list of filters, addressed by unique name. At the
     * moment it just use the built-in color tables, but at some point should be
     * enabled to read in from a extra gradient directory.
     * 
     * Since it does not support some sophisticated remove gui, for every layer
     * a new list is created, so that adding from the gui just adds it to the
     * current layer, while copying it to the gradient directory will make it
     * accessible everywhere.
     * 
     * It uses some old color tables which are included as vectors and scans the
     * directory Colortables for user defined additions.
     * 
     * @return Map of standard filters
     */
    public synchronized static Map<String, LUT> getStandardList() {
        if (standardList == null) {
            standardList = new TreeMap<String, LUT>();
            addStdLut("Gray", "ctable_idl_16bit_00_Black-White_Linear", -16777216, -16711423, -16645630, -16579837, -16514044, -16448251, -16382458, -16316665, -16250872, -16185079, -16119286, -16053493, -15987700, -15921907, -15856114, -15790321, -15724528, -15658735, -15592942, -15527149, -15461356, -15395563, -15329770, -15263977, -15198184, -15132391, -15066598, -15000805, -14935012, -14869219, -14803426, -14737633, -14671840, -14606047, -14540254, -14474461, -14408668, -14342875, -14277082, -14211289, -14145496, -14079703, -14013910, -13948117, -13882324, -13816531, -13750738, -13684945, -13619152, -13553359, -13487566, -13421773, -13355980, -13290187, -13224394, -13158601, -13092808, -13027015, -12961222, -12895429, -12829636, -12763843, -12698050, -12632257, -12566464, -12500671, -12434878, -12369085, -12303292, -12237499, -12171706, -12105913, -12040120, -11974327, -11908534, -11842741, -11776948, -11711155, -11645362, -11579569, -11513776, -11447983, -11382190, -11316397, -11250604, -11184811, -11119018, -11053225, -10987432, -10921639, -10855846, -10790053, -10724260, -10658467, -10592674, -10526881, -10461088, -10395295, -10329502, -10263709, -10197916, -10132123, -10066330, -10000537, -9934744, -9868951, -9803158, -9737365, -9671572, -9605779, -9539986, -9474193, -9408400, -9342607, -9276814, -9211021, -9145228, -9079435, -9013642, -8947849, -8882056, -8816263, -8750470, -8684677, -8618884, -8553091, -8487298, -8421505, -8355712, -8289919, -8224126, -8158333, -8092540, -8026747, -7960954, -7895161, -7829368, -7763575, -7697782, -7631989, -7566196, -7500403, -7434610, -7368817, -7303024, -7237231, -7171438, -7105645, -7039852, -6974059, -6908266, -6842473, -6776680, -6710887, -6645094, -6579301, -6513508, -6447715, -6381922, -6316129, -6250336, -6184543, -6118750, -6052957, -5987164, -5921371, -5855578, -5789785, -5723992, -5658199, -5592406, -5526613, -5460820, -5395027, -5329234, -5263441, -5197648, -5131855, -5066062, -5000269, -4934476, -4868683, -4802890, -4737097, -4671304, -4605511, -4539718, -4473925, -4408132, -4342339, -4276546, -4210753, -4144960, -4079167, -4013374, -3947581, -3881788, -3815995, -3750202, -3684409, -3618616, -3552823, -3487030, -3421237, -3355444, -3289651, -3223858, -3158065, -3092272, -3026479, -2960686, -2894893, -2829100, -2763307, -2697514, -2631721, -2565928, -2500135, -2434342, -2368549, -2302756, -2236963, -2171170, -2105377, -2039584, -1973791, -1907998, -1842205, -1776412, -1710619, -1644826, -1579033, -1513240, -1447447, -1381654, -1315861, -1250068, -1184275, -1118482, -1052689, -986896, -921103, -855310, -789517, -723724, -657931, -592138, -526345, -460552, -394759, -328966, -263173, -197380, -131587, -65794, -1);
            addStdLut("STEREO EUVI 171 " + angstrom, "ctable_idl_16bit_EUVI_171", -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777215, -16777212, -16777211, -16777210, -16777208, -16777208, -16777204, -16777202, -16777200, -16777198, -16777196, -16777195, -16777193, -16777191, -16777188, -16777185, -16777183, -16777181, -16777180, -16777179, -16777174, -16777173, -16777172, -16777170, -16777168, -16777166, -16777165, -16777164, -16777158, -16777157, -16777155, -16777155, -16777153, -16777151, -16777150, -16777143, -16777142, -16777141, -16777138, -16777138, -16777136, -16777135, -16777132, -16777128, -16777127, -16777126, -16777123, -16777122, -16777122, -16777119, -16777115, -16777113, -16777112, -16777111, -16777105, -16777104, -16777104, -16777100, -16777098, -16777097, -16777096, -16777092, -16777090, -16777089, -16777088, -16776573, -16776060, -16775802, -16775289, -16774005, -16773747, -16773234, -16771950, -16771693, -16771179, -16770666, -16769638, -16769125, -16768611, -16767583, -16767070, -16766556, -16766556, -16765786, -16765015, -16764502, -16763731, -16762960, -16762447, -16761676, -16761419, -16761419, -16760392, -16760135, -16758337, -16757567, -16757309, -16756796, -16756025, -16756025, -16755512, -16754741, -16753457, -16752944, -16752686, -16752173, -16751402, -16750889, -16750631, -16750118, -16748834, -16748577, -16748063, -16746779, -16746522, -16746009, -16745751, -16744467, -16743954, -16743696, -16743183, -16742412, -16741899, -16740615, -16739844, -16739844, -16739587, -16739073, -16738561, -16737793, -16146689, -15884289, -15096577, -15096577, -14833921, -14571521, -14046465, -13521153, -12996097, -12471041, -12208385, -11945985, -11683329, -11683329, -11158273, -10305025, -10042369, -9517313, -9254913, -8992257, -8729601, -7942145, -7416833, -7154433, -6629377, -6366721, -6104065, -5841665, -4988417, -4725761, -4463361, -4200705, -3675649, -3412993, -3150593, -2625537, -2100225, -1575169, -1575169, -1312769, -787457, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1);
            addStdLut("STEREO EUVI 284 " + angstrom, "ctable_idl_16bit_EUVI_284", -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16711424, -16579840, -16448512, -16382720, -16316928, -16119808, -16054016, -15856896, -15791104, -15659520, -15528192, -15462400, -15396608, -15199488, -15199488, -15067904, -14870784, -14739200, -14607872, -14542080, -14476288, -14213376, -14147584, -14147584, -14016256, -13950464, -13818880, -13621760, -13555968, -13292800, -13227008, -13095680, -13029888, -13029888, -12898304, -12766976, -12701184, -12504064, -12372480, -12175360, -12109568, -11977984, -11977984, -11846656, -11780864, -11583744, -11452160, -11386368, -11255040, -11189248, -10926336, -10926336, -10794752, -10663424, -10531840, -10466048, -10334720, -10268928, -10071552, -9939968, -9874176, -9874176, -9611264, -9545472, -9414144, -9348352, -9085440, -9019648, -8953856, -8822528, -8690944, -8690944, -8625152, -8428032, -8165120, -8099328, -8033536, -7902208, -7770624, -7704832, -7704832, -7573504, -7507712, -7244800, -7179008, -6981888, -6850304, -6718720, -6652928, -6652928, -6521344, -6324224, -6258432, -6192640, -6061312, -5929728, -5666816, -5601024, -5601024, -5403904, -5338112, -5272320, -5140992, -5009408, -4812288, -4746496, -4680704, -4483584, -4417792, -4352000, -4220672, -3957760, -3891968, -3826176, -3760384, -3563264, -3497472, -3365888, -3365888, -3037184, -2971392, -2905600, -2839808, -2642688, -2576896, -2445568, -2379776, -2379776, -2116864, -2051072, -1985280, -1722368, -1656576, -1525248, -1393664, -1196544, -1196544, -1130752, -1064960, -999168, -802048, -736256, -473344, -276224, -210432, -210432, -144640, -13056, -12793, -12010, -12010, -11746, -11227, -10963, -10444, -10444, -10180, -9397, -9397, -8878, -8614, -8095, -8095, -7831, -7048, -6529, -6529, -6265, -6002, -5482, -5482, -5219, -4436, -3916, -3916, -3916, -3653, -2870, -2870, -2087, -1567, -1304, -1304, -1304, -784, -521, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1);
            addStdLut("STEREO EUVI 195 " + angstrom, "ctable_idl_16bit_EUVI_195", -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16776960, -16776448, -16776192, -16775936, -16775680, -16775680, -16775168, -16774912, -16774400, -16774144, -16773632, -16773376, -16773120, -16772864, -16772352, -16772352, -16771840, -16771584, -16770816, -16770560, -16770304, -16770048, -16769280, -16769024, -16769024, -16768768, -16768512, -16768000, -16767744, -16767488, -16767232, -16766464, -16766208, -16765952, -16765696, -16765696, -16765184, -16764672, -16764160, -16763648, -16763392, -16763136, -16762880, -16762368, -16762368, -16762112, -16761856, -16761344, -16760832, -16760576, -16760320, -16760064, -16759552, -16759296, -16759296, -16758784, -16758016, -16757760, -16757504, -16757248, -16756736, -16756224, -16755968, -16755968, -16755712, -16755200, -16754944, -16754688, -16754432, -16753664, -16753408, -16753152, -16752896, -16752384, -16752384, -16555264, -16489472, -16161024, -16029696, -15963904, -15832576, -15635456, -15504128, -15504128, -15438336, -15307008, -14978560, -14847232, -14781440, -14650112, -14452992, -14321664, -14255872, -14255872, -13796096, -13730304, -13598976, -13467648, -13270528, -13204736, -12942080, -12876288, -12876288, -12679168, -12547840, -12416512, -12350720, -12088064, -11890944, -11825152, -11693824, -11496704, -11496704, -11299584, -11168256, -10839808, -10774016, -10642688, -10511360, -10314240, -10182912, -10117120, -10117120, -9985792, -9657344, -9591552, -9460224, -9394432, -9131776, -9065984, -8934656, -8803328, -8540416, -8409088, -8277760, -8211968, -8014848, -7883520, -7752189, -7555062, -7357935, -7357935, -7226604, -7160808, -7029477, -6832350, -6701019, -6503892, -6372560, -6175433, -6044102, -5978306, -5846975, -5649848, -5452721, -5321390, -5255594, -4992931, -4927136, -4795805, -4730009, -4730009, -4467346, -4270219, -4138888, -4073092, -3875966, -3744634, -3613303, -3547507, -3219049, -3087717, -3021922, -2890591, -2693464, -2562132, -2496337, -2365005, -2036547, -1970752, -1970752, -1839420, -1708089, -1510962, -1379630, -1313835, -1116708, -854045, -788250, -656918, -591123, -328460, -262664, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1);
            addStdLut("STEREO EUVI 304 " + angstrom, "ctable_idl_16bit_EUVI_304", -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16515072, -16449536, -16318464, -16252928, -16252928, -16056320, -15925248, -15663104, -15466496, -15400960, -15269888, -15204352, -15007744, -15007744, -14811136, -14614528, -14417920, -14352384, -14221312, -14155776, -13893632, -13762560, -13697024, -13565952, -13369344, -13303808, -13172736, -13107200, -12845056, -12648448, -12517376, -12517376, -12320768, -12255232, -12189696, -11993088, -11796480, -11599872, -11468800, -11272192, -11272192, -11206656, -11141120, -10944512, -10616832, -10551296, -10420224, -10289152, -10158080, -10158080, -9961472, -9699328, -9568256, -9502720, -9371648, -9240576, -9043968, -8912896, -8912896, -8650752, -8519680, -8454144, -8388608, -8060928, -7995392, -7864320, -7602176, -7602176, -7536640, -7405568, -7340032, -7012352, -6946816, -6750208, -6553600, -6488064, -6488064, -6356992, -6291456, -5963776, -5898240, -5701632, -5505024, -5439488, -5308160, -5308160, -5110528, -4912896, -4715264, -4649216, -4451840, -4385792, -4188160, -4056576, -4056576, -3792896, -3726848, -3595520, -3397888, -3200256, -3134208, -3002624, -2870528, -2739200, -2673152, -2541568, -2277888, -2146304, -2080256, -1948672, -1685248, -1619200, -1619200, -1487616, -1223936, -1092352, -1026560, -762880, -631296, -565248, -433664, -433664, -170240, -38656, -38144, -36608, -36096, -35584, -34816, -33792, -33792, -33280, -32256, -31229, -30713, -29937, -29421, -28389, -27873, -27357, -26841, -26065, -25034, -24518, -24002, -22970, -22454, -21678, -21678, -20646, -19614, -19098, -18583, -17551, -16775, -16259, -14711, -14711, -14195, -13679, -13163, -11872, -11356, -10840, -9292, -8776, -8776, -8516, -8000, -6452, -5937, -4905, -4129, -3613, -3097, -3097, -1549, -1033, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1);
            addStdLut("SOHO EIT 171 " + angstrom, "ctable_idl_16bit_EIT_171", -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777215, -16777212, -16777211, -16777210, -16777208, -16777208, -16777206, -16777204, -16777202, -16777200, -16777198, -16777196, -16777195, -16777193, -16777191, -16777191, -16777188, -16777187, -16777185, -16777183, -16777181, -16777180, -16777179, -16777174, -16777173, -16777173, -16777172, -16777170, -16777168, -16777166, -16777165, -16777164, -16777160, -16777158, -16777157, -16777155, -16777155, -16777153, -16777151, -16777150, -16777147, -16777145, -16777143, -16777142, -16777141, -16777138, -16777138, -16777136, -16777135, -16777132, -16777130, -16777128, -16777127, -16777126, -16777123, -16777122, -16777122, -16777119, -16777117, -16777115, -16777113, -16777112, -16777111, -16777108, -16777105, -16777104, -16777104, -16777103, -16777100, -16777098, -16777097, -16777096, -16777092, -16777090, -16777089, -16777088, -16776573, -16776573, -16776060, -16775802, -16775289, -16774005, -16773747, -16773234, -16772721, -16771950, -16771693, -16771693, -16771179, -16770666, -16769638, -16769125, -16768611, -16768354, -16767583, -16767070, -16766556, -16766556, -16765786, -16765015, -16764502, -16764244, -16763731, -16762960, -16762447, -16761676, -16761419, -16761419, -16760392, -16760135, -16759621, -16759364, -16758337, -16757567, -16757309, -16756796, -16756025, -16756025, -16755512, -16755254, -16754741, -16753457, -16752944, -16752686, -16752173, -16751402, -16750889, -16750631, -16750631, -16750118, -16748834, -16748577, -16748063, -16747806, -16746779, -16746522, -16746009, -16745751, -16745751, -16744467, -16743954, -16743696, -16743183, -16742412, -16741899, -16741642, -16740615, -16739844, -16739844, -16739587, -16739073, -16738561, -16737793, -16737537, -16146689, -15884289, -15359233, -15096577, -15096577, -14833921, -14571521, -14046465, -13521153, -13258753, -12996097, -12471041, -12208385, -11945985, -11683329, -11683329, -11158273, -10567681, -10305025, -10042369, -9517313, -9254913, -8992257, -8729601, -7942145, -7942145, -7679489, -7416833, -7154433, -6629377, -6366721, -6104065, -5841665, -4988417, -4725761, -4725761, -4463361, -4200705, -3675649, -3412993, -3150593, -2625537, -2100225, -1837825, -1575169, -1575169, -1312769, -787457, -525057, -1);
            addStdLut("SOHO EIT 284 " + angstrom, "ctable_idl_16bit_EIT_284", -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16711424, -16579840, -16448512, -16382720, -16316928, -16316928, -16119808, -16054016, -15856896, -15791104, -15659520, -15528192, -15462400, -15396608, -15199488, -15199488, -15067904, -14936576, -14870784, -14739200, -14607872, -14542080, -14476288, -14213376, -14147584, -14147584, -14016256, -13950464, -13818880, -13687552, -13621760, -13555968, -13292800, -13227008, -13095680, -13029888, -13029888, -12898304, -12766976, -12701184, -12504064, -12372480, -12306688, -12175360, -12109568, -11977984, -11977984, -11846656, -11780864, -11583744, -11452160, -11386368, -11255040, -11189248, -11057664, -10926336, -10926336, -10794752, -10663424, -10531840, -10466048, -10334720, -10268928, -10071552, -9939968, -9874176, -9874176, -9742848, -9611264, -9545472, -9414144, -9348352, -9085440, -9019648, -8953856, -8822528, -8690944, -8690944, -8625152, -8493824, -8428032, -8165120, -8099328, -8033536, -7902208, -7770624, -7704832, -7704832, -7573504, -7507712, -7244800, -7179008, -7113216, -6981888, -6850304, -6718720, -6652928, -6652928, -6521344, -6324224, -6258432, -6192640, -6061312, -5929728, -5798400, -5666816, -5601024, -5601024, -5403904, -5338112, -5272320, -5140992, -5009408, -4812288, -4746496, -4680704, -4483584, -4483584, -4417792, -4352000, -4220672, -3957760, -3891968, -3826176, -3760384, -3563264, -3497472, -3365888, -3365888, -3300096, -3037184, -2971392, -2905600, -2839808, -2642688, -2576896, -2445568, -2379776, -2379776, -2116864, -2051072, -1985280, -1919488, -1722368, -1656576, -1525248, -1393664, -1196544, -1196544, -1130752, -1064960, -999168, -802048, -736256, -539136, -473344, -276224, -210432, -210432, -144640, -13056, -12793, -12010, -12010, -11746, -11227, -10963, -10963, -10444, -10444, -10180, -9397, -9397, -8878, -8614, -8095, -8095, -7831, -7048, -7048, -6529, -6529, -6265, -6002, -5482, -5482, -5219, -4436, -3916, -3916, -3916, -3653, -3133, -2870, -2870, -2087, -1567, -1304, -1304, -1304, -784, -521, -1, -1);
            addStdLut("SOHO EIT 195 " + angstrom, "ctable_idl_16bit_EIT_195", -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16776960, -16776448, -16776192, -16775936, -16775680, -16775680, -16775168, -16774912, -16774400, -16774144, -16773632, -16773376, -16773120, -16772864, -16772352, -16772352, -16771840, -16771584, -16771328, -16770816, -16770560, -16770304, -16770048, -16769280, -16769024, -16769024, -16768768, -16768512, -16768000, -16767744, -16767488, -16767232, -16766464, -16766208, -16765952, -16765696, -16765696, -16765184, -16764928, -16764672, -16764160, -16763648, -16763392, -16763136, -16762880, -16762368, -16762368, -16762112, -16761856, -16761344, -16760832, -16760576, -16760320, -16760064, -16759552, -16759296, -16759296, -16758784, -16758528, -16758016, -16757760, -16757504, -16757248, -16756736, -16756224, -16755968, -16755968, -16755712, -16755200, -16754944, -16754688, -16754432, -16753664, -16753408, -16753152, -16752896, -16752384, -16752384, -16686592, -16555264, -16489472, -16161024, -16029696, -15963904, -15832576, -15635456, -15504128, -15504128, -15438336, -15307008, -14978560, -14847232, -14781440, -14650112, -14452992, -14321664, -14255872, -14255872, -14058752, -13796096, -13730304, -13598976, -13467648, -13270528, -13204736, -12942080, -12876288, -12876288, -12679168, -12547840, -12416512, -12350720, -12088064, -11890944, -11825152, -11693824, -11496704, -11496704, -11365376, -11299584, -11168256, -10839808, -10774016, -10642688, -10511360, -10314240, -10182912, -10117120, -10117120, -9985792, -9657344, -9591552, -9460224, -9394432, -9131776, -9065984, -8934656, -8803328, -8803328, -8540416, -8409088, -8277760, -8211968, -8014848, -7883520, -7752189, -7555062, -7357935, -7357935, -7226604, -7160808, -7029477, -6832350, -6701019, -6503892, -6372560, -6175433, -6044102, -6044102, -5978306, -5846975, -5649848, -5452721, -5321390, -5255594, -4992931, -4927136, -4795805, -4730009, -4730009, -4467346, -4270219, -4138888, -4073092, -3875966, -3744634, -3613303, -3547507, -3219049, -3219049, -3087717, -3021922, -2890591, -2693464, -2562132, -2496337, -2365005, -2036547, -1970752, -1970752, -1839420, -1708089, -1510962, -1379630, -1313835, -1116708, -854045, -788250, -656918, -656918, -591123, -328460, -262664, -1);
            addStdLut("SOHO EIT 304 " + angstrom, "ctable_idl_16bit_EIT_304", -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16777216, -16711680, -16515072, -16449536, -16318464, -16252928, -16252928, -16056320, -15925248, -15794176, -15663104, -15466496, -15400960, -15269888, -15204352, -15007744, -15007744, -14811136, -14745600, -14614528, -14417920, -14352384, -14221312, -14155776, -13893632, -13762560, -13762560, -13697024, -13565952, -13369344, -13303808, -13172736, -13107200, -12845056, -12713984, -12648448, -12517376, -12517376, -12320768, -12255232, -12189696, -11993088, -11796480, -11665408, -11599872, -11468800, -11272192, -11272192, -11206656, -11141120, -10944512, -10747904, -10616832, -10551296, -10420224, -10289152, -10158080, -10158080, -9961472, -9895936, -9699328, -9568256, -9502720, -9371648, -9240576, -9043968, -8912896, -8912896, -8847360, -8650752, -8519680, -8454144, -8388608, -8060928, -7995392, -7864320, -7798784, -7602176, -7602176, -7536640, -7405568, -7340032, -7012352, -6946816, -6815744, -6750208, -6553600, -6488064, -6488064, -6356992, -6291456, -5963776, -5898240, -5767168, -5701632, -5505024, -5439488, -5308160, -5308160, -5110528, -4912896, -4846848, -4715264, -4649216, -4451840, -4385792, -4188160, -4056576, -4056576, -3858944, -3792896, -3726848, -3595520, -3397888, -3200256, -3134208, -3002624, -2870528, -2870528, -2739200, -2673152, -2541568, -2277888, -2146304, -2080256, -1948672, -1816832, -1685248, -1619200, -1619200, -1487616, -1223936, -1092352, -1026560, -960512, -762880, -631296, -565248, -433664, -433664, -170240, -38656, -38144, -37632, -36608, -36096, -35584, -34816, -33792, -33792, -33280, -32768, -32256, -31229, -30713, -29937, -29421, -28389, -27873, -27873, -27357, -26841, -26065, -25034, -24518, -24002, -22970, -22454, -21938, -21678, -21678, -20646, -19614, -19098, -18583, -17551, -17291, -16775, -16259, -14711, -14711, -14195, -13679, -13163, -12388, -11872, -11356, -10840, -9292, -8776, -8776, -8516, -8000, -6968, -6452, -5937, -4905, -4129, -3613, -3097, -3097, -2581, -1549, -1033, -1);
            addStdLut("Blue/White Linear", "ctable_idl_16bit_01_Blue-White_Linear", -16777216, -16777215, -16777214, -16777212, -16777211, -16777210, -16777208, -16777207, -16777206, -16777204, -16777203, -16777202, -16777200, -16777199, -16777198, -16777196, -16777195, -16777193, -16777192, -16777191, -16777189, -16777188, -16777187, -16777185, -16777184, -16777183, -16777181, -16777180, -16777179, -16777177, -16777176, -16777174, -16777173, -16777172, -16777170, -16777169, -16777168, -16777166, -16777165, -16777164, -16777162, -16777161, -16777160, -16777158, -16777157, -16777155, -16777154, -16777153, -16777151, -16777150, -16777149, -16777147, -16777146, -16777145, -16777143, -16777142, -16777141, -16777139, -16777138, -16777136, -16777135, -16777134, -16777132, -16777131, -16777130, -16777128, -16777127, -16777126, -16777124, -16777123, -16777122, -16777120, -16777119, -16777117, -16777116, -16777115, -16777113, -16777112, -16777111, -16777109, -16777108, -16777107, -16777105, -16777104, -16777103, -16777101, -16777100, -16777098, -16777097, -16777096, -16777094, -16777093, -16777092, -16777090, -16777089, -16777088, -16777086, -16776573, -16776060, -16775802, -16775289, -16774776, -16774518, -16774005, -16773747, -16773234, -16772721, -16772463, -16771950, -16771693, -16771179, -16770666, -16770409, -16769895, -16769638, -16769125, -16768611, -16768354, -16767840, -16767583, -16767070, -16766556, -16766299, -16765786, -16765528, -16765015, -16764502, -16764244, -16763731, -16763474, -16762960, -16762447, -16762189, -16761676, -16761419, -16760905, -16760392, -16760135, -16759621, -16759364, -16758851, -16758337, -16758080, -16757567, -16757309, -16756796, -16756282, -16756025, -16755512, -16755254, -16754741, -16754228, -16753970, -16753457, -16752944, -16752686, -16752173, -16751916, -16751402, -16750889, -16750631, -16750118, -16749861, -16749347, -16748834, -16748577, -16748063, -16747806, -16747293, -16746779, -16746522, -16746009, -16745751, -16745238, -16744724, -16744467, -16743954, -16743696, -16743183, -16742670, -16742412, -16741899, -16741642, -16741128, -16740615, -16740358, -16739844, -16739587, -16739073, -16738561, -16738305, -16737793, -16737537, -16409345, -16146689, -15884289, -15621633, -15359233, -15096577, -14833921, -14571521, -14308865, -14046465, -13783809, -13521153, -13258753, -12996097, -12733441, -12471041, -12208385, -11945985, -11683329, -11420673, -11158273, -10830081, -10567681, -10305025, -10042369, -9779969, -9517313, -9254913, -8992257, -8729601, -8467201, -8204545, -7942145, -7679489, -7416833, -7154433, -6891777, -6629377, -6366721, -6104065, -5841665, -5579009, -5251073, -4988417, -4725761, -4463361, -4200705, -3938305, -3675649, -3412993, -3150593, -2887937, -2625537, -2362881, -2100225, -1837825, -1575169, -1312769, -1050113, -787457, -525057, -262401, -1);
            addStdLut("Red Temperature", "ctable_idl_16bit_03_Red_Temperature", -16777216, -16711680, -16646144, -16515072, -16449536, -16318464, -16252928, -16121856, -16056320, -15925248, -15859712, -15794176, -15663104, -15597568, -15466496, -15400960, -15269888, -15204352, -15073280, -15007744, -14942208, -14811136, -14745600, -14614528, -14548992, -14417920, -14352384, -14221312, -14155776, -14024704, -13959168, -13893632, -13762560, -13697024, -13565952, -13500416, -13369344, -13303808, -13172736, -13107200, -13041664, -12910592, -12845056, -12713984, -12648448, -12517376, -12451840, -12320768, -12255232, -12189696, -12058624, -11993088, -11862016, -11796480, -11665408, -11599872, -11468800, -11403264, -11272192, -11206656, -11141120, -11010048, -10944512, -10813440, -10747904, -10616832, -10551296, -10420224, -10354688, -10289152, -10158080, -10092544, -9961472, -9895936, -9764864, -9699328, -9568256, -9502720, -9371648, -9306112, -9240576, -9109504, -9043968, -8912896, -8847360, -8716288, -8650752, -8519680, -8454144, -8388608, -8257536, -8192000, -8060928, -7995392, -7864320, -7798784, -7667712, -7602176, -7536640, -7405568, -7340032, -7208960, -7143424, -7012352, -6946816, -6815744, -6750208, -6619136, -6553600, -6488064, -6356992, -6291456, -6160384, -6094848, -5963776, -5898240, -5767168, -5701632, -5636096, -5505024, -5439488, -5308160, -5242112, -5110528, -5044480, -4912896, -4846848, -4715264, -4649216, -4583168, -4451840, -4385792, -4254208, -4188160, -4056576, -3990528, -3858944, -3792896, -3726848, -3595520, -3529472, -3397888, -3331840, -3200256, -3134208, -3002624, -2936576, -2870528, -2739200, -2673152, -2541568, -2475520, -2343936, -2277888, -2146304, -2080256, -1948672, -1882880, -1816832, -1685248, -1619200, -1487616, -1421568, -1289984, -1223936, -1092352, -1026560, -960512, -828928, -762880, -631296, -565248, -433664, -367616, -236032, -170240, -38656, -38144, -37632, -37120, -36608, -36096, -35584, -35072, -34816, -34304, -33792, -33280, -32768, -32256, -31744, -31229, -30713, -30453, -29937, -29421, -28905, -28389, -27873, -27357, -26841, -26325, -26065, -25549, -25034, -24518, -24002, -23486, -22970, -22454, -21938, -21678, -21162, -20646, -20130, -19614, -19098, -18583, -18067, -17551, -17291, -16775, -16259, -15743, -15227, -14711, -14195, -13679, -13163, -12903, -12388, -11872, -11356, -10840, -10324, -9808, -9292, -8776, -8516, -8000, -7484, -6968, -6452, -5937, -5421, -4905, -4389, -4129, -3613, -3097, -2581, -2065, -1549, -1033, -517, -1);
            addStdLut("Blue/Green/Red/Yellow", "ctable_idl_16bit_04_Blue-Green-Red-Yellow", -16777216, -16777214, -16777212, -16777210, -16777208, -16777206, -16777204, -16777202, -16777200, -16777198, -16777196, -16777194, -16777191, -16777189, -16777187, -16777185, -16777183, -16777181, -16777179, -16777177, -16777175, -16777173, -16777171, -16777169, -16777166, -16777164, -16777162, -16777160, -16777158, -16777156, -16777154, -16777152, -16777150, -16776380, -16775610, -16774840, -16774069, -16773299, -16772529, -16771759, -16770733, -16769963, -16769193, -16768423, -16767653, -16766883, -16766113, -16765343, -16764316, -16763548, -16762780, -16762012, -16761244, -16760476, -16759708, -16758940, -16757916, -16757148, -16756380, -16755612, -16754844, -16754076, -16753308, -16752540, -16751516, -16750748, -16749980, -16749212, -16748444, -16747676, -16746908, -16746140, -16745116, -16744348, -16743580, -16742812, -16742044, -16741276, -16740508, -16739740, -16738716, -16738720, -16738723, -16738726, -16738729, -16738732, -16738735, -16738738, -16738741, -16738745, -16738748, -16738751, -16738754, -16738757, -16738760, -16738763, -16738766, -16739026, -16739285, -16739288, -16739547, -16739806, -16739809, -16740068, -16740071, -16740331, -16740590, -16740593, -16740852, -16741111, -16741114, -16741373, -16741376, -16283392, -15759616, -15301632, -14777856, -14319872, -13796096, -13338112, -12814336, -12356352, -11832576, -11374592, -10850816, -10392832, -9869056, -9411072, -8887296, -8561408, -8235264, -7909120, -7582976, -7257088, -6930944, -6604800, -6278656, -5952768, -5626624, -5300480, -4974336, -4648448, -4322304, -3996160, -3670016, -3669504, -3603456, -3602944, -3536640, -3536128, -3470080, -3469312, -3403264, -3402752, -3336448, -3335936, -3269888, -3269376, -3203072, -3202560, -3136512, -3135744, -3069696, -3069184, -3002880, -3002368, -2936320, -2935552, -2869504, -2868992, -2802944, -2802176, -2736128, -2735616, -2669312, -2668800, -2602752, -2601984, -2535936, -2535424, -2469120, -2468608, -2402560, -2402048, -2335744, -2335232, -2269184, -2268416, -2202368, -2201856, -2135552, -2135040, -2068992, -2068224, -2002176, -2001664, -1935616, -1934848, -1868800, -1868288, -1801984, -1801472, -1735424, -1734656, -1668608, -1668096, -1602048, -1601280, -1535232, -1534720, -1468416, -1467904, -1401856, -1401088, -1335040, -1334528, -1268224, -1267712, -1201664, -1201152, -1134848, -1134336, -1068288, -1067520, -1001472, -1000960, -934656, -934144, -868096, -867328, -801280, -800768, -734720, -733952, -667904, -667392, -601088, -600576, -534528, -533760, -467712, -467200, -400896, -400384, -334336, -333824, -267520, -267008, -200960, -200192, -134144, -133632, -67328, -66816, -768, -256);
            addStdLut("Green/White Linear", "ctable_idl_16bit_08_Green-White_Linear", -16777216, -16776960, -16776704, -16776448, -16776192, -16775936, -16775680, -16775424, -16775168, -16774912, -16774656, -16774400, -16774144, -16773888, -16773632, -16773376, -16773120, -16772864, -16772608, -16772352, -16772096, -16771840, -16771584, -16771328, -16771072, -16770816, -16770560, -16770304, -16770048, -16769792, -16769536, -16769280, -16769024, -16768768, -16768512, -16768256, -16768000, -16767744, -16767488, -16767232, -16766976, -16766720, -16766464, -16766208, -16765952, -16765696, -16765440, -16765184, -16764928, -16764672, -16764416, -16764160, -16763904, -16763648, -16763392, -16763136, -16762880, -16762624, -16762368, -16762112, -16761856, -16761600, -16761344, -16761088, -16760832, -16760576, -16760320, -16760064, -16759808, -16759552, -16759296, -16759040, -16758784, -16758528, -16758272, -16758016, -16757760, -16757504, -16757248, -16756992, -16756736, -16756480, -16756224, -16755968, -16755712, -16755456, -16755200, -16754944, -16754688, -16754432, -16754176, -16753920, -16753664, -16753408, -16753152, -16752896, -16752640, -16752384, -16686592, -16555264, -16489472, -16358144, -16226816, -16161024, -16029696, -15963904, -15832576, -15701248, -15635456, -15504128, -15438336, -15307008, -15175680, -15109888, -14978560, -14847232, -14781440, -14650112, -14584320, -14452992, -14321664, -14255872, -14124544, -14058752, -13927424, -13796096, -13730304, -13598976, -13467648, -13401856, -13270528, -13204736, -13073408, -12942080, -12876288, -12744960, -12679168, -12547840, -12416512, -12350720, -12219392, -12088064, -12022272, -11890944, -11825152, -11693824, -11562496, -11496704, -11365376, -11299584, -11168256, -11036928, -10971136, -10839808, -10774016, -10642688, -10511360, -10445568, -10314240, -10182912, -10117120, -9985792, -9920000, -9788672, -9657344, -9591552, -9460224, -9394432, -9263104, -9131776, -9065984, -8934656, -8803328, -8737536, -8606208, -8540416, -8409088, -8277760, -8211968, -8080640, -8014848, -7883520, -7752189, -7686394, -7555062, -7423731, -7357935, -7226604, -7160808, -7029477, -6898145, -6832350, -6701019, -6635223, -6503892, -6372560, -6306765, -6175433, -6044102, -5978306, -5846975, -5781180, -5649848, -5518517, -5452721, -5321390, -5255594, -5124263, -4992931, -4927136, -4795805, -4730009, -4598678, -4467346, -4401551, -4270219, -4138888, -4073092, -3941761, -3875966, -3744634, -3613303, -3547507, -3416176, -3350380, -3219049, -3087717, -3021922, -2890591, -2759259, -2693464, -2562132, -2496337, -2365005, -2233674, -2167878, -2036547, -1970752, -1839420, -1708089, -1642293, -1510962, -1379630, -1313835, -1182503, -1116708, -985377, -854045, -788250, -656918, -591123, -459791, -328460, -262664, -131333, -1);
            addStdLut("Green/White Exponential", "ctable_idl_16bit_09_Green-White_Exponential", -16711680, -16711680, -16711424, -16645632, -16645631, -16645375, -16579583, -16579583, -16513790, -16513534, -16513534, -16447742, -16447485, -16447485, -16381693, -16381437, -16315900, -16315644, -16315388, -16249851, -16249595, -16183803, -16183803, -16183546, -16117754, -16117754, -16117498, -16051705, -16051705, -15985913, -15985657, -15985656, -15919864, -15919608, -15919608, -15853815, -15853559, -15788023, -15787766, -15787510, -15721974, -15721718, -15655925, -15655925, -15655669, -15589877, -15589876, -15589620, -15523828, -15523828, -15458035, -15457779, -15457779, -15391987, -15391730, -15391730, -15325938, -15325681, -15260145, -15259889, -15259633, -15194096, -15193840, -15128048, -15128048, -15127791, -15061999, -15061999, -15061743, -14995950, -14995950, -14930158, -14929902, -14929901, -14864109, -14863853, -14863852, -14798060, -14797804, -14732268, -14732011, -14731755, -14666219, -14665963, -14600170, -14600170, -14599914, -14534122, -14534121, -14533865, -14468073, -14468073, -14402280, -14402024, -14402024, -14336231, -14335975, -14335975, -14270183, -14269926, -14204390, -14204134, -14203878, -14138341, -14138085, -14072293, -14072293, -14072036, -14006244, -14006244, -14005988, -13940195, -13940195, -13874403, -13874146, -13874146, -13808354, -13808098, -13808097, -13742305, -13742049, -13676513, -13676256, -13676000, -13610464, -13610208, -13544415, -13544415, -13544159, -13478367, -13478110, -13477854, -13411806, -13411549, -13345757, -13345245, -13344989, -13213660, -13213148, -13147356, -13081564, -13081051, -13015259, -12949467, -12948955, -12883162, -12817370, -12816858, -12751066, -12685273, -12684761, -12618969, -12553176, -12552664, -12421336, -12290008, -12158423, -12092631, -11961303, -11829719, -11698390, -11567062, -11435478, -11304150, -11172821, -11106773, -10975445, -10844117, -10712532, -10581204, -10449876, -10318291, -10252499, -10120915, -9989587, -9858258, -9726674, -9595346, -9464018, -9332433, -9266641, -9135313, -9003729, -8872400, -8741073, -8609487, -8478156, -8412361, -8280774, -8149443, -8018112, -7886525, -7755194, -7623863, -7492276, -7426481, -7295150, -7163563, -7032232, -6900902, -6769315, -6637984, -6506653, -6440602, -6309271, -6177940, -6046353, -5915022, -5783691, -5652104, -5586309, -5454978, -5323391, -5192061, -5060474, -4929143, -4797812, -4666225, -4600430, -4469099, -4337512, -4206181, -4074850, -3943263, -3811932, -3746137, -3614550, -3483220, -3351889, -3220302, -3088971, -2957640, -2826053, -2760258, -2628927, -2497340, -2366009, -2234678, -2103091, -1971760, -1905965, -1774379, -1643048, -1511717, -1380130, -1248799, -1117468, -985881, -920086, -788755, -657168, -525837, -394506, -262919, -131588, -1);
            addStdLut("Rainbow 1", "ctable_idl_16bit_13_Rainbow_1", -16777216, -16515069, -16187385, -15925238, -15597554, -15335405, -15007721, -14745572, -14417888, -14155738, -13828053, -13500368, -13238219, -12976069, -12779457, -12582844, -12320696, -12255155, -12058543, -11927466, -11730853, -11599777, -11534236, -11403160, -11337619, -11206543, -11272074, -11140998, -11075457, -11009916, -11140984, -11075443, -11075439, -11075434, -11206502, -11272033, -11272029, -11272024, -11337555, -11599695, -11665226, -11730758, -11796289, -12123965, -12189496, -12320564, -12451631, -12844842, -12975910, -13172513, -13303581, -13762328, -13958932, -14155535, -14417675, -14614278, -15138561, -15400705, -15728385, -15990529, -16514817, -16776961, -16775937, -16774913, -16772865, -16771585, -16770561, -16769537, -16767233, -16766209, -16765185, -16763905, -16762881, -16760833, -16759809, -16758529, -16757505, -16755457, -16754177, -16753153, -16752129, -16749825, -16748801, -16747777, -16746497, -16744449, -16743425, -16742401, -16741121, -16740097, -16738049, -16736769, -16735745, -16734721, -16732417, -16731393, -16730369, -16729089, -16727041, -16726017, -16724993, -16723713, -16721665, -16720641, -16719361, -16718337, -16717313, -16715009, -16713985, -16712961, -16711681, -16711690, -16711694, -16711698, -16711703, -16711711, -16711716, -16711720, -16711724, -16711733, -16711737, -16711741, -16711745, -16711749, -16711758, -16711762, -16711766, -16711771, -16711779, -16711784, -16711788, -16711792, -16711801, -16711805, -16711809, -16711813, -16711822, -16711826, -16711830, -16711834, -16711839, -16711847, -16711852, -16711856, -16711860, -16711869, -16711873, -16711877, -16711881, -16711890, -16711894, -16711898, -16711902, -16711911, -16711915, -16711920, -16711924, -16711928, -16711936, -16449792, -16187648, -15925504, -15335680, -15073536, -14811392, -14549248, -13959424, -13697280, -13369600, -13107456, -12583168, -12321024, -11993344, -11731200, -11469056, -10879232, -10617088, -10354944, -10092800, -9502976, -9240832, -8913152, -8651008, -8126720, -7864576, -7536896, -7274752, -6684928, -6422784, -6160640, -5898496, -5636352, -5046528, -4784384, -4456704, -4194560, -3670272, -3408128, -3080448, -2818304, -2228480, -1966336, -1704192, -1442048, -852224, -590080, -327936, -256, -1536, -3584, -4608, -5888, -6912, -8960, -10240, -11264, -12288, -14592, -15616, -16640, -17664, -19968, -20992, -22016, -23296, -24320, -26368, -27648, -28672, -29696, -32000, -33024, -34048, -35072, -37376, -38400, -39424, -40704, -42752, -43776, -45056, -46080, -47104, -49408, -50432, -51456, -52480, -54784, -55808, -56832, -58112, -60160, -61184, -62464, -63488, -65536);
            addStdLut("Blue/Red", "ctable_idl_16bit_33_Blue-Red", -16777085, -16777085, -16777081, -16777077, -16777073, -16777069, -16777065, -16777061, -16777057, -16777053, -16777049, -16777045, -16777041, -16777037, -16777033, -16777029, -16777025, -16777021, -16777017, -16777013, -16777009, -16777005, -16777001, -16776997, -16776993, -16776989, -16776985, -16776981, -16776977, -16776973, -16776969, -16776965, -16776961, -16776961, -16776193, -16775169, -16774145, -16773121, -16772097, -16771073, -16770049, -16769025, -16768001, -16766977, -16765953, -16764929, -16763905, -16762881, -16761857, -16760833, -16759809, -16758785, -16757761, -16756737, -16755713, -16754689, -16753665, -16752641, -16751617, -16750593, -16749569, -16748545, -16747521, -16746497, -16745473, -16744449, -16743425, -16742401, -16741377, -16740353, -16739329, -16738305, -16737281, -16736257, -16735233, -16734209, -16733185, -16732161, -16731137, -16730113, -16729089, -16728065, -16727041, -16726017, -16724993, -16723969, -16722945, -16721921, -16720897, -16719873, -16718849, -16717825, -16716801, -16715777, -16714753, -16713729, -16712705, -16711681, -16711681, -16515077, -16252937, -15990797, -15728657, -15466517, -15204377, -14942237, -14680097, -14417957, -14155817, -13893677, -13631537, -13369397, -13107257, -12845117, -12582977, -12320837, -12058697, -11796557, -11534417, -11272277, -11010137, -10747997, -10485857, -10223717, -9961577, -9699437, -9437297, -9175157, -8913017, -8650877, -8388737, -8126597, -7864457, -7602317, -7340177, -7078037, -6815897, -6553757, -6291617, -6029477, -5767337, -5505197, -5243057, -4980917, -4718777, -4456637, -4194497, -3932357, -3670217, -3408077, -3145937, -2883797, -2621657, -2359517, -2097377, -1835237, -1573097, -1310957, -1048817, -786677, -524537, -262397, -256, -1280, -2304, -3328, -4352, -5376, -6400, -7424, -8448, -9472, -10496, -11520, -12544, -13568, -14592, -15616, -16640, -17664, -18688, -19712, -20736, -21760, -22784, -23808, -24832, -25856, -26880, -27904, -28928, -29952, -30976, -32000, -33024, -34048, -35072, -36096, -37120, -38144, -39168, -40192, -41216, -42240, -43264, -44288, -45312, -46336, -47360, -48384, -49408, -50432, -51456, -52480, -53504, -54528, -55552, -56576, -57600, -58624, -59648, -60672, -61696, -62720, -63744, -64768, -65536, -393216, -655360, -983040, -1245184, -1507328, -1835008, -2097152, -2424832, -2686976, -2949120, -3276800, -3538944, -3866624, -4128768, -4390912, -4718592, -4980736, -5308416, -5570560, -5832704, -6160384, -6422528, -6750208, -7012352, -7274496, -7602176, -7864320, -8192000, -8192000);
            addStdLut("Rainbow 2", "ctable_idl_16bit_34_Rainbow_2", -8650497, -8650497, -8912641, -9240321, -9502465, -9830145, -10092289, -10419969, -10682113, -11009793, -11271937, -11599617, -11861761, -12189441, -12451585, -12779265, -13041409, -13369089, -13631233, -13958913, -14221057, -14548737, -14810881, -15138561, -15400705, -15728385, -15990529, -16318209, -16580353, -16776449, -16775425, -16774145, -16773121, -16771841, -16770817, -16769537, -16768513, -16767233, -16766209, -16764929, -16763905, -16762625, -16761601, -16760321, -16759297, -16758017, -16756993, -16755713, -16754689, -16753409, -16752385, -16751105, -16750081, -16748801, -16747777, -16746497, -16745473, -16744193, -16743169, -16741889, -16740865, -16739585, -16738561, -16737281, -16736257, -16734977, -16733953, -16732673, -16731649, -16730369, -16729345, -16728065, -16727041, -16725761, -16724737, -16723457, -16722433, -16721153, -16720129, -16718849, -16717825, -16716545, -16715265, -16714241, -16712961, -16711937, -16711685, -16711689, -16711694, -16711698, -16711703, -16711707, -16711712, -16711716, -16711721, -16711725, -16711730, -16711734, -16711739, -16711743, -16711748, -16711752, -16711757, -16711761, -16711766, -16711770, -16711775, -16711779, -16711784, -16711788, -16711793, -16711797, -16711802, -16711806, -16711811, -16711815, -16711820, -16711824, -16711829, -16711833, -16711838, -16711842, -16711847, -16711851, -16711856, -16711860, -16711865, -16711869, -16711874, -16711878, -16711883, -16711887, -16711892, -16711896, -16711901, -16711905, -16711910, -16711914, -16711919, -16711923, -16711928, -16711932, -16646400, -16384256, -15859968, -15794432, -15466752, -15204608, -14876928, -14614784, -14287104, -14024960, -13697280, -13435136, -13107456, -12845312, -12517632, -12255488, -11927808, -11665664, -11337984, -11075840, -10748160, -10486016, -10158336, -9896192, -9568512, -9306368, -8978688, -8651008, -8388864, -8061184, -7799040, -7471360, -7209216, -6881536, -6619392, -6291712, -6029568, -5701888, -5439744, -5112064, -4849920, -4522240, -4260096, -3932416, -3670272, -3342592, -3080448, -2752768, -2490624, -2162944, -1900800, -1573120, -1310976, -983296, -721152, -393472, -131328, -1024, -2048, -3328, -4352, -5632, -6656, -7936, -8960, -10240, -11264, -12544, -13568, -14848, -15872, -17152, -18176, -19456, -20480, -21760, -22784, -24064, -25088, -26368, -27392, -28672, -29696, -30976, -32000, -33280, -34304, -35584, -36608, -37888, -38912, -40192, -41216, -42496, -43520, -44800, -45824, -47104, -48128, -49408, -50432, -51712, -52736, -54016, -55040, -56320, -57344, -58624, -59648, -60928, -61952, -63232, -64256, -64256);
            // From the resources
            String ggrFiles[] = { "AIA94", "AIA131", "AIA171", "AIA193", "AIA211", "AIA304", "AIA335", "AIA1600", "AIA1700", "AIA4500" };
            for (String file : ggrFiles) {
                try {
                    LUT l = readGimpGradientStream(FileUtils.getResourceInputStream("/ggr/" + file + ".ggr"));
                    standardList.put(l.getName(), l);
                } catch (FileNotFoundException e) {
                    Log.warn("Could not restore gimp gradient file " + file, e);
                } catch (GradientError e) {
                    Log.warn("Could not restore gimp gradient file " + file, e);
                } catch (IOException e) {
                    Log.warn("Could not restore gimp gradient file " + file, e);
                }
            }
            // User addons
            File addOnDir = JHVDirectory.COLOR_PLUGINS.getFile();
            for (File f : addOnDir.listFiles(new GGRFilter())) {
                // Try to add f
                try {
                    LUT l = readGimpGradientFile(f);
                    standardList.put(l.getName(), l);
                } catch (FileNotFoundException e) {
                    Log.warn("Error loading color table plugin dir", e);
                } catch (GradientError e) {
                    Log.warn("Error loading color table plugin dir", e);
                } catch (IOException e) {
                    Log.warn("Error loading color table plugin dir", e);
                }
            }
        }
        return new TreeMap<String, LUT>(standardList);
    }

    /**
     * Shortcut to create the old standard gradients. Better way is to use
     * readGimpGradient properly
     * 
     * @param name
     *            Name to identify
     * @param lookup8
     *            8 bit lookup values
     * @param fname
     *            Filename for 16 bit
     */
    private static void addStdLut(String name, String fname, int... lookup8) {
        LUT l = new LUT(name, lookup8, internal16Readin(fname));
        standardList.put(name, l);
    }

    /**
     * Reads in the existing 16bit LUT. In general a gimp gradient file is much
     * more common and should be used for additions
     * 
     * @param name
     * @return 16bit lookup table
     */
    private static int[] internal16Readin(String name) {
        try {
            BufferedReader br = new BufferedReader(new InputStreamReader(FileUtils.getResourceInputStream("/luts/" + name)));
            String strLine;

            int[] intArray = new int[65536];
            int counter = 0;
            while ((strLine = br.readLine()) != null) {

                String[] array = strLine.split(" ");

                for (int i = 0; i < array.length; i++) {
                    if (array[i].equals(" ") || array[i].equals("")) {

                    } else {
                        intArray[counter] = Integer.parseInt(array[i]);
                        counter++;
                    }
                }
            }
            br.close();
            return intArray;
        } catch (Exception e) {// Catch exception if any
            Log.error("Error open internal color table " + name, e);
            return null;
        }
    }
}
